/*
**                    dfttest v1.1 for Avisynth 2.5.x
**
**   2D/3D frequency domain denoiser.
**
**   Copyright (C) 2007 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "dfttest.h"

PVideoFrame __stdcall dfttest::GetFrame(int n, IScriptEnvironment *env)
{
	if (tbsize == 1)
		return GetFrame_S(n, env);
	return GetFrame_T(n, env);
}

unsigned __stdcall threadPool(void *ps)
{
	const PS_INFO *pss = (PS_INFO*)ps;
	while (true)
	{
		WaitForSingleObject(pss->nextJob,INFINITE);
		if (pss->type < 0)
			return 0;
		if (pss->type < 2)
			func_0(ps);
		else
			func_1(ps);
		ResetEvent(pss->nextJob);
		SetEvent(pss->jobFinished);
	}
}

void proc0_C(const unsigned char *s0, const float *s1, float *d,
	const int p0, const int p1)
{
	for (int u=0; u<p1; ++u)
	{
		for (int v=0; v<p1; ++v)
			d[v] = s0[v]*s1[v];
		s0 += p0;
		s1 += p1;
		d += p1;
	}
}

void proc0_SSE(const unsigned char *s0, const float *s1, float *d,
	const int p0, const int p1)
{
	__asm
	{
		mov esi,s0
		mov edi,s1
		mov edx,d
		mov eax,p1
		pxor mm7,mm7
uloop:
		mov ecx,p1
		xor ebx,ebx
vloop:
		movd mm0,[esi+ebx]
		punpcklbw mm0,mm7
		movq mm1,mm0
		punpcklbw mm0,mm7
		punpckhbw mm1,mm7
		cvtpi2ps xmm0,mm0
		cvtpi2ps xmm1,mm1
		shufps xmm0,xmm1,68
		mulps xmm0,[edi+ebx*4]
		movaps [edx+ebx*4],xmm0
		add ebx,4
		sub ecx,4
		jnz vloop
		add esi,p0
		mov ecx,p1
		lea edi,[edi+ecx*4]
		lea edx,[edx+ecx*4]
		sub eax,1
		jnz uloop
		emms
	}
}

void proc1_C(const float *s0, const float *s1, float *d,
	const int p0, const int p1)
{
	for (int u=0; u<p0; ++u)
	{
		for (int v=0; v<p0; ++v)
			d[v] += s0[v]*s1[v];
		s0 += p0;
		s1 += p0;
		d += p1;
	}
}

void proc1_SSE(const float *s0, const float *s1, float *d,
	const int p0, const int p1)
{
	__asm
	{
		mov esi,s0
		mov edi,s1
		mov edx,d
		mov eax,p0
uloop:
		mov ecx,p0
		xor ebx,ebx
vloop:
		movaps xmm0,[esi+ebx*4]
		mulps xmm0,[edi+ebx*4]
		movups xmm1,[edx+ebx*4]
		addps xmm0,xmm1
		movups [edx+ebx*4],xmm0
		add ebx,4
		sub ecx,4
		jnz vloop
		mov ecx,p0
		lea esi,[esi+ecx*4]
		lea edi,[edi+ecx*4]
		mov ecx,p1
		lea edx,[edx+ecx*4]
		sub eax,1
		jnz uloop
	}
}

void func_0(void *ps)
{
	PS_INFO *pss = (PS_INFO*)ps;
	const int b = pss->b;
	const int sheight = pss->sheight[b];
	const int eheight = pss->eheight[b];
	const int width = pss->pf->GetWidth(b);
	const int src_pitch = pss->pf->GetPitch(b);
	const unsigned char *srcp = pss->pf->GetPtr(b)+sheight*src_pitch;
	float *ebp_saved = pss->ebuff[b]+width*sheight;
	const int sbsize = pss->sbsize;
	const int sbd1 = pss->sbsize>>1;
	const int sosize = pss->sosize;
	float *dftr = pss->dftr;
	const float *hw = pss->hw;
	fftwf_complex *dftc = pss->dftc;
	const int inc = (pss->type&1) ? sbsize-sosize : 1;
	for (int y=sheight; y<eheight; y+=inc)
	{
		for (int x=0; x<=width-sbsize; x+=inc)
		{
			proc0(srcp+x,hw,dftr,src_pitch,sbsize);
			pss->fftwf_execute_dft_r2c(pss->ft,dftr,dftc);
			filterCoeffs((float*)dftc,(float*)pss->dftgc,pss->sigmas,pss->ccnt);
			pss->fftwf_execute_dft_c2r(pss->fti,dftc,dftr);
			if (pss->type&1)
				proc1(dftr,hw,ebp_saved+x,sbsize,width);
			else
				ebp_saved[x+sbd1*width+sbd1] = 
					dftr[sbd1*sbsize+sbd1]*hw[sbd1*sbsize+sbd1];
		}
		srcp += src_pitch*inc;
		ebp_saved += width*inc;
	}
}

void func_1(void *ps)
{
	PS_INFO *pss = (PS_INFO*)ps;
	nlCache *fc = pss->fc;
	const int b = pss->b;
	PlanarFrame *src = fc->frames[0]->ppf;
	const int sheight = pss->sheight[b];
	const int eheight = pss->eheight[b];
	const int width = src->GetWidth(b);
	const int src_pitch = src->GetPitch(b);
	float *ebp_saved = pss->ebuff[b];
	const int sbsize = pss->sbsize;
	const int sbd1 = pss->sbsize>>1;
	const int sosize = pss->sosize;
	const int stopz = pss->stopz;
	float *dftr = pss->dftr;
	const float *hw = pss->hw;
	fftwf_complex *dftc = pss->dftc;
	const unsigned char **pfplut = pss->pfplut;
	for (int i=0; i<fc->size; ++i)
		pfplut[i] = fc->frames[fc->getCachePos(i)]->ppf->GetPtr(b)+src_pitch*sheight;
	const int inc = (pss->type&1) ? sbsize-sosize : 1;
	for (int y=sheight; y<eheight; y+=inc)
	{
		for (int x=0; x<=width-sbsize; x+=inc)
		{
			for (int z=0; z<=stopz; ++z)
				proc0(pfplut[z]+x,hw+sbsize*sbsize*z,
					dftr+sbsize*sbsize*z,src_pitch,sbsize);
			pss->fftwf_execute_dft_r2c(pss->ft,dftr,dftc);
			filterCoeffs((float*)dftc,(float*)pss->dftgc,pss->sigmas,pss->ccnt);
			pss->fftwf_execute_dft_c2r(pss->fti,dftc,dftr);
			float *ebpT = ebp_saved+y*width+x;
			const float *dftrTT = dftr+pss->pos*pss->barea;
			const float *hwTT = hw+pss->pos*pss->barea;
			if (pss->type&1)
				proc1(dftrTT,hwTT,ebpT,sbsize,width);
			else
				ebpT[sbd1*width+sbd1] = 
					dftrTT[sbd1*sbsize+sbd1]*hwTT[sbd1*sbsize+sbd1];
		}
		for (int q=0; q<fc->size; ++q)
			pfplut[q] += src_pitch*inc;
	}
}

__forceinline int intcast(float f)
{
	int i;
	_asm fld f
	_asm fistp i
	return i;
}

PVideoFrame dfttest::GetFrame_S(int n, IScriptEnvironment *env)
{
	nlf->pf->copyFrom(child->GetFrame(mapn(n), env), vi);
	copyPad(nlf->pf,nlf->ppf,env);
	for (int b=0; b<3; ++b)
	{
		if ((b == 0 && !Y) || (b == 1 && !U) || (b == 2 && !V))
		{
			nlf->pf->copyPlaneTo(*dstPF, b);
			continue;
		}
		const int height = nlf->ppf->GetHeight(b);
		const int width = nlf->ppf->GetWidth(b);
		memset(ebuff[b],0,height*width*sizeof(float));
		for (int i=0; i<threads; ++i)
		{
			pssInfo[i]->b = b;
			ResetEvent(pssInfo[i]->jobFinished);
			SetEvent(pssInfo[i]->nextJob);
		}
		for (int i=0; i<threads; ++i)
			WaitForSingleObject(pssInfo[i]->jobFinished,INFINITE);
		unsigned char *dstp = dstPF->GetPtr(b);
		const int dst_pitch = dstPF->GetPitch(b);
		const int src_width = dstPF->GetWidth(b);
		const int src_height = dstPF->GetHeight(b);
		float *ebp = ebuff[b]+width*((height-src_height)>>1)+((width-src_width)>>1);
		unsigned int savedfp = _control87(_RC_CHOP,_MCW_RC);
		for (int y=0; y<src_height; ++y)
		{
			for (int x=0; x<src_width; ++x)
				dstp[x] = min(max(intcast(ebp[x]+0.5f),0),255);
			dstp += dst_pitch;
			ebp += width;
		}
		_control87(savedfp,_MCW_RC);
	}
	PVideoFrame dst = env->NewVideoFrame(vi);
	dstPF->copyTo(dst, vi);
	return dst;
}

PVideoFrame dfttest::GetFrame_T(int n, IScriptEnvironment *env)
{
	for (int b=0; b<3; ++b)
	{
		if ((b == 0 && !Y) || (b == 1 && !U) || (b == 2 && !V))
			continue;
		PlanarFrame *src = fc->frames[0]->ppf;
		memset(ebuff[b],0,src->GetHeight(b)*src->GetWidth(b)*sizeof(float));
	}
	int pos;
	if (tmode == 0)
	{
		pos = tbsize>>1;
		fc->resetCacheStart(n-pos,n+pos);
		for (int i=n-pos; i<=n+pos; ++i)
		{
			nlFrame *nl = fc->frames[fc->getCachePos(i-n+pos)];
			if (nl->fnum != i)
			{
				nl->pf->copyFrom(child->GetFrame(mapn2(i), env), vi);
				copyPad(nl->pf,nl->ppf,env);
				nl->setFNum(i);
			}
		}
		processTemporalBlock(pos);
	}
	else
	{
		pos = 0;
		for (int z=-max((tbsize-tosize),tosize); z<=n; z+=(tbsize-tosize))
		{
			if (n < z+tbsize)
			{
				pos = n-z;
				fc->resetCacheStart(z,z+tbsize-1);
				for (int i=z; i<=z+tbsize-1; ++i)
				{
					nlFrame *nl = fc->frames[fc->getCachePos(i-z)];
					if (nl->fnum != i)
					{
						nl->pf->copyFrom(child->GetFrame(mapn2(i), env), vi);
						copyPad(nl->pf,nl->ppf,env);
						nl->setFNum(i);
					}
				}
				processTemporalBlock(pos);
			}
		}
	}
	for (int b=0; b<3; ++b)
	{
		if ((b == 0 && !Y) || (b == 1 && !U) || (b == 2 && !V))
		{
			fc->frames[fc->getCachePos(pos)]->pf->copyPlaneTo(*dstPF,b);
			continue;
		}
		PlanarFrame *src = fc->frames[0]->ppf;
		const int height = src->GetHeight(b);
		const int width = src->GetWidth(b);
		unsigned char *dstp = dstPF->GetPtr(b);
		const int dst_pitch = dstPF->GetPitch(b);
		const int src_width = dstPF->GetWidth(b);
		const int src_height = dstPF->GetHeight(b);
		float *ebp = ebuff[b]+width*((height-src_height)>>1)+((width-src_width)>>1);
		unsigned int savedfp = _control87(_RC_CHOP,_MCW_RC);
		for (int y=0; y<src_height; ++y)
		{
			for (int x=0; x<src_width; ++x)
				dstp[x] = min(max(intcast(ebp[x]+0.5f),0),255);
			dstp += dst_pitch;
			ebp += width;
		}
		_control87(savedfp,_MCW_RC);
	}
	PVideoFrame dst = env->NewVideoFrame(vi);
	dstPF->copyTo(dst, vi);
	return dst;
}

void dfttest::processTemporalBlock(int pos)
{
	for (int i=0; i<threads; ++i)
		pssInfo[i]->pos = pos;
	for (int b=0; b<3; ++b)
	{
		if ((b == 0 && !Y) || (b == 1 && !U) || (b == 2 && !V))
			continue;
		for (int i=0; i<threads; ++i)
		{
			pssInfo[i]->b = b;
			ResetEvent(pssInfo[i]->jobFinished);
			SetEvent(pssInfo[i]->nextJob);
		}
		for (int i=0; i<threads; ++i)
			WaitForSingleObject(pssInfo[i]->jobFinished,INFINITE);
	}
}

void filter_0_ZM_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	const float gf = dftc[0]/dftgc[0];
	for (int h=0; h<ccnt; h+=2)
	{
		const float g0 = gf*dftgc[h+0];
		const float c0 = dftc[h+0]-g0;
		const float g1 = gf*dftgc[h+1];
		const float c1 = dftc[h+1]-g1;
		const float psd = c0*c0+c1*c1+1e-15f;
		const float coeff = max((psd-sigmas[h])/psd,0.0f);
		dftc[h+0] = c0*coeff+g0;
		dftc[h+1] = c1*coeff+g1;
	}
}

void filter_0_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	for (int h=0; h<ccnt; h+=2)
	{
		const float psd = dftc[h+0]*dftc[h+0]+dftc[h+1]*dftc[h+1]+1e-15f;
		const float coeff = max((psd-sigmas[h])/psd,0.0f);
		dftc[h+0] *= coeff;
		dftc[h+1] *= coeff;
	}
}

void filter_1_ZM_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	const float gf = dftc[0]/dftgc[0];
	for (int h=0; h<ccnt; h+=2)
	{
		const float g0 = gf*dftgc[h+0];
		const float c0 = dftc[h+0]-g0;
		const float g1 = gf*dftgc[h+1];
		const float c1 = dftc[h+1]-g1;
		const float psd = c0*c0+c1*c1;
		if (psd < sigmas[h])
		{
			dftc[h+0] = g0;
			dftc[h+1] = g1;
		}
	}
}

void filter_1_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	for (int h=0; h<ccnt; h+=2)
	{
		const float psd = dftc[h+0]*dftc[h+0]+dftc[h+1]*dftc[h+1];
		if (psd < sigmas[h])
			dftc[h+0] = dftc[h+1] = 0.0f;
	}
}

void filter_2_ZM_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	const float gf = dftc[0]/dftgc[0];
	for (int h=0; h<ccnt; h+=2)
	{
		const float g0 = gf*dftgc[h+0];
		const float g1 = gf*dftgc[h+1];
		dftc[h+0] = (dftc[h+0]-g0)*sigmas[h]+g0;
		dftc[h+1] = (dftc[h+1]-g1)*sigmas[h]+g1;
	}
}

void filter_2_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	for (int h=0; h<ccnt; h+=2)
	{
		dftc[h+0] *= sigmas[h];
		dftc[h+1] *= sigmas[h];
	}
}

void filter_0_ZM_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	__asm
	{
		mov esi,dftgc
		mov edi,dftc
		mov edx,sigmas
		mov ecx,ccnt
		xor eax,eax
		movss xmm6,[edi]
		xorps xmm5,xmm5
		divss xmm6,[esi]
		movaps xmm7,sse_1em15
		shufps xmm6,xmm6,0
four_loop:
		movaps xmm0,[esi+eax*4]
		mulps xmm0,xmm6
		movaps xmm1,[edi+eax*4]
		subps xmm1,xmm0
		movaps xmm2,xmm1
		mulps xmm2,xmm2
		movaps xmm3,xmm2
		shufps xmm3,xmm3,177
		addps xmm2,xmm7
		addps xmm3,xmm2
		movaps xmm2,xmm3
		subps xmm3,[edx+eax*4]
		divps xmm3,xmm2
		maxps xmm3,xmm5
		mulps xmm1,xmm3
		addps xmm1,xmm0
		movaps [edi+eax*4],xmm1
		add eax,4
		cmp eax,ecx
		jl four_loop
	}
}

void filter_0_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	__asm
	{
		mov edi,dftc
		mov edx,sigmas
		mov ecx,ccnt
		xor eax,eax
		xorps xmm5,xmm5
		movaps xmm7,sse_1em15
four_loop:
		movaps xmm1,[edi+eax*4]
		movaps xmm2,xmm1
		mulps xmm2,xmm2
		movaps xmm3,xmm2
		shufps xmm3,xmm3,177
		addps xmm2,xmm7
		addps xmm3,xmm2
		movaps xmm2,xmm3
		subps xmm3,[edx+eax*4]
		divps xmm3,xmm2
		maxps xmm3,xmm5
		mulps xmm1,xmm3
		movaps [edi+eax*4],xmm1
		add eax,4
		cmp eax,ecx
		jl four_loop
	}
}

void filter_1_ZM_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	__asm
	{
		mov esi,dftgc
		mov edi,dftc
		mov edx,sigmas
		mov ecx,ccnt
		xor eax,eax
		movss xmm6,[edi]
		divss xmm6,[esi]
		shufps xmm6,xmm6,0
four_loop:
		movaps xmm0,[esi+eax*4]
		mulps xmm0,xmm6
		movaps xmm1,[edi+eax*4]
		subps xmm1,xmm0
		movaps xmm2,xmm1
		mulps xmm2,xmm2
		movaps xmm3,xmm2
		shufps xmm3,xmm3,177
		addps xmm3,xmm2
		movaps xmm4,[edx+eax*4]
		cmpps xmm4,xmm3,2
		andps xmm1,xmm4
		addps xmm1,xmm0
		movaps [edi+eax*4],xmm1
		add eax,4
		cmp eax,ecx
		jl four_loop
	}
}

void filter_1_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	__asm
	{
		mov edi,dftc
		mov edx,sigmas
		mov ecx,ccnt
		xor eax,eax
four_loop:
		movaps xmm1,[edi+eax*4]
		movaps xmm2,xmm1
		mulps xmm2,xmm2
		movaps xmm3,xmm2
		shufps xmm3,xmm3,177
		addps xmm3,xmm2
		movaps xmm4,[edx+eax*4]
		cmpps xmm4,xmm3,2
		andps xmm1,xmm4
		movaps [edi+eax*4],xmm1
		add eax,4
		cmp eax,ecx
		jl four_loop
	}
}

void filter_2_ZM_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	__asm
	{
		mov esi,dftgc
		mov edi,dftc
		mov edx,sigmas
		mov ecx,ccnt
		xor eax,eax
		movss xmm6,[edi]
		divss xmm6,[esi]
		shufps xmm6,xmm6,0
four_loop:
		movaps xmm0,[esi+eax*4]
		mulps xmm0,xmm6
		movaps xmm1,[edi+eax*4]
		subps xmm1,xmm0
		mulps xmm1,[edx+eax*4]
		addps xmm1,xmm0
		movaps [edi+eax*4],xmm1
		add eax,4
		cmp eax,ecx
		jl four_loop
	}
}

void filter_2_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt)
{
	__asm
	{
		mov edi,dftc
		mov edx,sigmas
		mov ecx,ccnt
		xor eax,eax
four_loop:
		movaps xmm0,[edi+eax*4]
		mulps xmm0,[edx+eax*4]
		movaps [edi+eax*4],xmm0
		add eax,4
		cmp eax,ecx
		jl four_loop
	}
}

void dfttest::copyPad(PlanarFrame *src, PlanarFrame *dst, IScriptEnvironment *env)
{
	for (int b=0; b<3; ++b)
	{
		if ((b == 0 && !Y) || (b == 1 && !U) || (b == 2 && !V))
			continue;
		const int src_width = src->GetWidth(b);
		const int src_height = src->GetHeight(b);
		const int src_pitch = src->GetPitch(b);
		const int dst_width = dst->GetWidth(b);
		const int dst_height = dst->GetHeight(b);
		const int dst_pitch = dst->GetPitch(b);
		const int offy = (dst_height-src_height)>>1;
		const int offx = (dst_width-src_width)>>1;
		env->BitBlt(dst->GetPtr(b)+dst->GetPitch(b)*offy+offx,dst->GetPitch(b),
			src->GetPtr(b),src->GetPitch(b),src->GetWidth(b),src->GetHeight(b));
		unsigned char *dstp = dst->GetPtr(b)+dst->GetPitch(b)*offy;
		for (int y=offy; y<src_height+offy; ++y)
		{
			int w = offx*2;
			for (int x=0; x<offx; ++x, --w)
				dstp[x] = dstp[w];
			w = offx+src_width-2;
			for (int x=offx+src_width; x<dst_width; ++x, --w)
				dstp[x] = dstp[w];
			dstp += dst_pitch;
		}
		int w = offy*2;
		for (int y=0; y<offy; ++y, --w)
		{
			env->BitBlt(dst->GetPtr(b)+dst->GetPitch(b)*y,dst->GetPitch(b),
				dst->GetPtr(b)+dst->GetPitch(b)*w,dst->GetPitch(b),dst->GetWidth(b),1);
		}
		w = offy+src_height-2;
		for (int y=offy+src_height; y<dst_height; ++y, --w)
		{
			env->BitBlt(dst->GetPtr(b)+dst->GetPitch(b)*y,dst->GetPitch(b),
				dst->GetPtr(b)+dst->GetPitch(b)*w,dst->GetPitch(b),dst->GetWidth(b),1);
		}
	}
}

int dfttest::mapn(int n)
{
	if (n < 0) return 0;
	if (n >= vi.num_frames) return vi.num_frames-1;
	return n;
}

int dfttest::mapn2(int n)
{
	if (n < 0)
	{
		if (0-n < vi.num_frames)
			return 0-n;
		return 0;
	}
	if (n >= vi.num_frames)
	{
		const int nf = vi.num_frames-1;
		if (2*nf-n >= 0)
			return 2*nf-n;
		return vi.num_frames-1;
	}
	return n;
}

nlFrame::nlFrame()
{
	fnum = -20;
	pf = ppf = NULL;
}

nlFrame::nlFrame(PlanarFrame *tp, VideoInfo &vi)
{
	fnum = -20;
	pf = new PlanarFrame(vi);
	ppf = new PlanarFrame();
	ppf->createFromPlanar(*tp);
}

nlFrame::~nlFrame()
{
	if (pf) delete pf;
	if (ppf) delete ppf;
}

void nlFrame::setFNum(int i)
{
	fnum = i;
}

nlCache::nlCache()
{
	frames = NULL;
	start_pos = size = -20;
}

nlCache::nlCache(int _size, PlanarFrame *tp, VideoInfo &vi)
{
	frames = NULL;
	start_pos = size = -20;
	if (_size > 0)
	{
		start_pos = 0;
		size = _size;
		frames = (nlFrame**)malloc(size*sizeof(nlFrame*));
		memset(frames, 0, size*sizeof(nlFrame*));
		for (int i=0; i<size; ++i)
			frames[i] = new nlFrame(tp, vi);
	}
}

nlCache::~nlCache()
{
	if (frames)
	{
		for (int i=0; i<size; ++i)
		{
			if (frames[i]) 
				delete frames[i];
		}
		free(frames);
	}
}

void nlCache::resetCacheStart(int first, int last)
{
	for (int j=first; j<=last; ++j)
	{
		for (int i=0; i<size; ++i)
		{
			if (frames[i]->fnum == j)
			{
				start_pos = i-j+first;
				if (start_pos < 0) start_pos += size;
				else if (start_pos >= size) start_pos -= size;
				return;
			}
		}
	}
}

int nlCache::getCachePos(int n)
{
	return (start_pos+n)%size;
}

dfttest::dfttest(PClip _child, bool _Y, bool _U, bool _V, int _ftype, float _sigma, 
	int _sbsize, int _smode, int _sosize, int _tbsize, int _tmode, int _tosize, 
	int _swin, int _twin, double _sbeta, double _tbeta, bool _zmean, 
	const char* _cfile, int _threads, int _opt, IScriptEnvironment *env) : 
	GenericVideoFilter(_child), Y(_Y), U(_U), V(_V), ftype(_ftype), sigma(_sigma),
	sbsize(_sbsize), smode(_smode), sosize(_sosize), tbsize(_tbsize), tmode(_tmode),
	tosize(_tosize), swin(_swin), twin(_twin), sbeta(_sbeta), tbeta(_tbeta), 
	zmean(_zmean), cfile(_cfile), opt(_opt), threads(_threads)
{
	ft = fti = ftg = NULL;
	ebuff = NULL;
	hw = sigmas = NULL;
	dftgc = NULL;
	dstPF = NULL;
	fc = NULL;
	nlf = NULL;
	hLib = NULL;
	if (!vi.IsYUY2() && !vi.IsYV12())
		env->ThrowError("dfttest:  only yv12 and yuy2 input supported!");
	if (ftype < 0 || ftype > 2)
		env->ThrowError("dfttest:  ftype must be set to 0, 1, or 2!");
	if (tbsize < 1)
		env->ThrowError("dfttest:  tbsize must be >= 1!");
	if (tmode != 0 && tmode != 1)
		env->ThrowError("dfttest:  tmode must be set to 0 or 1!");
	if (!(tbsize&1) && tmode == 0)
		env->ThrowError("dfttest:  tbsize must be odd when using tmode=0!");
	if (tmode == 0)
		tosize = 0;
	if (tosize < 0)
		env->ThrowError("dfttest:  tosize must be >= 0!");
	if (tosize >= tbsize)
		env->ThrowError("dfttest:  tosize must be < tbsize!");
	if (tbsize > vi.num_frames)
		env->ThrowError("dfttest:  tbsize must be less than or equal to the " \
			"number of frames in the clip!");
	if (tosize > (tbsize>>1) && (tbsize%(tbsize-tosize)) != 0)
		env->ThrowError("dfttest:  temporal overlap greater than 50%c " \
			"requires that (tbsize-tosize) is a divisor of tbsize!", '%');
	if (twin < 0 || twin > 11)
		env->ThrowError("dfttest:  twin must be between 0 and 11 (inclusive)!");
	if (sbsize < 1)
		env->ThrowError("dfttest:  sbsize must be >= 1!");
	if (smode != 0 && smode != 1)
		env->ThrowError("dfttest:  smode must be set to 0 or 1!");
	if (!(sbsize&1) && smode == 0)
		env->ThrowError("dfttest:  sbsize must be odd when using smode=0!");
	if (smode == 0)
		sosize = 0;
	if (sosize < 0)
		env->ThrowError("dfttest:  sosize must be >= 0!");
	if (sosize >= sbsize)
		env->ThrowError("dfttest:  sosize must be < sbsize!");
	if (sosize > (sbsize>>1) && (sbsize%(sbsize-sosize)) != 0)
		env->ThrowError("dfttest:  spatial overlap greater than 50%c " \
			"requires that (sbsize-sosize) is a divisor of sbsize!", '%');
	if (swin < 0 || swin > 11)
		env->ThrowError("dfttest:  swin must be between 0 and 11 (inclusive)!");
	if (threads < 0 || threads > 16)
		env->ThrowError("dfttest:  threads must be between 0 and 16 (inclusive)!");
	if (opt < 0 || opt > 2)
		env->ThrowError("dfttest:  opt must be set to 0, 1, or 2!");
	if (threads == 0)
		threads = num_processors();
	hLib = LoadLibrary("libfftw3f-3.dll");
	if (hLib)
	{
		fftwf_destroy_plan = 
			(fftwf_destroy_plan_proc)GetProcAddress(hLib, "fftwf_destroy_plan");
		fftwf_plan_dft_r2c_3d = 
			(fftwf_plan_dft_r2c_3d_proc)GetProcAddress(hLib, "fftwf_plan_dft_r2c_3d");
		fftwf_plan_dft_c2r_3d = 
			(fftwf_plan_dft_c2r_3d_proc)GetProcAddress(hLib, "fftwf_plan_dft_c2r_3d");
		fftwf_plan_dft_r2c_2d = 
			(fftwf_plan_dft_r2c_2d_proc)GetProcAddress(hLib, "fftwf_plan_dft_r2c_2d");
		fftwf_plan_dft_c2r_2d = 
			(fftwf_plan_dft_c2r_2d_proc)GetProcAddress(hLib, "fftwf_plan_dft_c2r_2d");
		fftwf_execute_dft_r2c = 
			(fftwf_execute_dft_r2c_proc)GetProcAddress(hLib, "fftwf_execute_dft_r2c");
		fftwf_execute_dft_c2r = 
			(fftwf_execute_dft_c2r_proc)GetProcAddress(hLib, "fftwf_execute_dft_c2r");
	}
	if (!hLib || !fftwf_destroy_plan || !fftwf_execute_dft_r2c || 
		!fftwf_execute_dft_c2r || !fftwf_plan_dft_r2c_3d || !fftwf_plan_dft_c2r_3d || 
		!fftwf_plan_dft_r2c_2d || !fftwf_plan_dft_c2r_2d)
		env->ThrowError("dfttest:  unable to load libfftw3f-3.dll!");
	child->SetCacheHints(CACHE_NOTHING,0);
	barea = sbsize*sbsize;
	bvolume = barea*tbsize;
	norm = 1.0f/float(bvolume);
	ccnt = ((sbsize>>1)+1)*sbsize*tbsize;
	if (((env->GetCPUFlags()&CPUF_SSE) && opt == 0) || opt == 2)
	{
		if (!(sbsize&3))
		{
			proc0 = proc0_SSE;
			proc1 = proc1_SSE;
		}
		else
		{
			proc0 = proc0_C;
			proc1 = proc1_C;
		}
		if (zmean)
		{
			if (ftype == 0) filterCoeffs = filter_0_ZM_SSE;
			else if (ftype == 1) filterCoeffs = filter_1_ZM_SSE;
			else filterCoeffs = filter_2_ZM_SSE;
		}
		else
		{
			if (ftype == 0) filterCoeffs = filter_0_SSE;
			else if (ftype == 1) filterCoeffs = filter_1_SSE;
			else filterCoeffs = filter_2_SSE;
		}
	}
	else
	{
		proc0 = proc0_C;
		proc1 = proc1_C;
		if (zmean)
		{
			if (ftype == 0) filterCoeffs = filter_0_ZM_C;
			else if (ftype == 1) filterCoeffs = filter_1_ZM_C;
			else filterCoeffs = filter_2_ZM_C;
		}
		else
		{
			if (ftype == 0) filterCoeffs = filter_0_C;
			else if (ftype == 1) filterCoeffs = filter_1_C;
			else filterCoeffs = filter_2_C;
		}
	}
	if (smode == 0)
	{
		const int ae = (sbsize>>1)<<1;
		noxl = vi.width+ae;
		noyl = vi.height+ae;
		noxc = (vi.width>>1)+ae;
		if (vi.IsYV12()) noyc = (vi.height>>1)+ae;
		else noyc = vi.height+ae;
	}
	else
	{
		const int ae = max(sbsize-sosize,sosize)*2;
		noxl = vi.width+EXTRA(vi.width,sbsize)+ae;
		noyl = vi.height+EXTRA(vi.height,sbsize)+ae;
		noxc = (vi.width>>1)+EXTRA(vi.width>>1,sbsize)+ae;
		if (vi.IsYV12()) noyc = (vi.height>>1)+EXTRA(vi.height>>1,sbsize)+ae;
		else noyc = vi.height+EXTRA(vi.height,sbsize)+ae;
	}
	PlanarFrame *padPF = new PlanarFrame();
	padPF->createPlanar(noyl,noyc,noxl,noxc);
	if (tbsize > 1)
		fc = new nlCache(tbsize,padPF,(VideoInfo)vi);
	else
		nlf = new nlFrame(padPF,(VideoInfo)vi);
	ebuff = (float**)calloc(3,sizeof(float*));
	for (int b=0; b<3; ++b)
	{
		if ((b == 0 && !Y) || (b == 1 && !U) || (b == 2 && !V))
			continue;
		ebuff[b] = (float*)_aligned_malloc(padPF->GetWidth(b)*
			padPF->GetHeight(b)*sizeof(float),16);
	}
	delete padPF;
	dstPF = new PlanarFrame(vi);
	int w = 0;
	hw = (float*)_aligned_malloc(bvolume*sizeof(float),16);
	for (int j=0; j<tbsize; ++j)
	{
		const double dt = getWVal(j,tbsize,tosize,twin,tbeta);
		for (int i=0; i<sbsize; ++i)
		{
			const double ht = getWVal(i,sbsize,sosize,swin,sbeta);
			for (int k=0; k<sbsize; ++k, ++w)
			{
				const double wt = getWVal(k,sbsize,sosize,swin,sbeta);
				hw[w] = (float)(dt*ht*wt);
			}
		}
	}
	normalizeWindow();
	float *dftgr = (float*)_aligned_malloc(bvolume*sizeof(float),16);
	dftgc = (fftwf_complex*)_aligned_malloc(sizeof(fftwf_complex)*(ccnt+3),16);
	if (tbsize > 1)
		ftg = fftwf_plan_dft_r2c_3d(tbsize, sbsize, sbsize, dftgr, dftgc, FFTW_PATIENT|FFTW_DESTROY_INPUT);
	else
		ftg = fftwf_plan_dft_r2c_2d(sbsize, sbsize, dftgr, dftgc, FFTW_PATIENT|FFTW_DESTROY_INPUT);
	const float *hwT = hw;
	float *dftgrT = dftgr;
	for (int s=0; s<tbsize; ++s)
	{
		for (int i=0; i<sbsize; ++i)
		{
			for (int k=0; k<sbsize; ++k)
				dftgrT[k] = 255.0f*hwT[k];
			dftgrT += sbsize;
			hwT += sbsize;
		}
	}
	fftwf_execute_dft_r2c(ftg,dftgr,dftgc);
	sigmas = (float*)_aligned_malloc((ccnt*2+2)*sizeof(float),16);
	if (*cfile)
		loadCFile(env);
	else
	{
		for (int i=0; i<ccnt*2; ++i)
			sigmas[i] = sigma*sigma;
	}
	fftwf_complex *ta = (fftwf_complex*)_aligned_malloc(sizeof(fftwf_complex)*(ccnt+3),16);
	if (tbsize > 1)
	{
		ft = fftwf_plan_dft_r2c_3d(tbsize, sbsize, sbsize, dftgr, ta, FFTW_PATIENT|FFTW_DESTROY_INPUT);
		fti = fftwf_plan_dft_c2r_3d(tbsize, sbsize, sbsize, ta, dftgr, FFTW_PATIENT|FFTW_DESTROY_INPUT);
	}
	else
	{
		ft = fftwf_plan_dft_r2c_2d(sbsize, sbsize, dftgr, ta, FFTW_PATIENT|FFTW_DESTROY_INPUT);
		fti = fftwf_plan_dft_c2r_2d(sbsize, sbsize, ta, dftgr, FFTW_PATIENT|FFTW_DESTROY_INPUT);
	}
	_aligned_free(ta);
	_aligned_free(dftgr);
	tids = (unsigned*)malloc(threads*sizeof(unsigned));
	thds = (HANDLE*)malloc(threads*sizeof(HANDLE));
	pssInfo = (PS_INFO**)malloc(threads*sizeof(PS_INFO*));
	PlanarFrame *pf = tbsize <= 1 ? nlf->ppf : fc->frames[0]->ppf;
	for (int i=0; i<threads; ++i)
	{
		pssInfo[i] = (PS_INFO*)calloc(1,sizeof(PS_INFO));
		pssInfo[i]->type = (tbsize>1?2:0)+smode;
		pssInfo[i]->ccnt = ccnt<<1;
		pssInfo[i]->stopz = tbsize-1;
		pssInfo[i]->barea = barea;
		pssInfo[i]->sbsize = sbsize;
		pssInfo[i]->sosize = sosize;
		pssInfo[i]->hw = hw;
		pssInfo[i]->sigmas = sigmas;
		pssInfo[i]->ebuff = ebuff;
		pssInfo[i]->dftgc = dftgc;
		pssInfo[i]->ft = ft;
		pssInfo[i]->fti = fti;
		if (tbsize <= 1)
			pssInfo[i]->pf = nlf->ppf;
		else
			pssInfo[i]->pfplut = (const unsigned char **)_aligned_malloc(
				fc->size*sizeof(const unsigned char*),16);
		pssInfo[i]->fc = fc;
		pssInfo[i]->fftwf_execute_dft_r2c = fftwf_execute_dft_r2c;
		pssInfo[i]->fftwf_execute_dft_c2r = fftwf_execute_dft_c2r;
		pssInfo[i]->dftr = (float*)_aligned_malloc(bvolume*sizeof(float),16);
		pssInfo[i]->dftc = (fftwf_complex*)_aligned_malloc(sizeof(fftwf_complex)*(ccnt+3),16);
		for (int b=0; b<3; ++b)
		{
			const int height = smode == 0 ? pf->GetHeight(b)-((sbsize>>1)<<1) : 
				(pf->GetHeight(b)-sosize)/(sbsize-sosize);
			const int hslice = height/threads;
			const int hremain = height%threads;
			if (smode == 0)
			{
				pssInfo[i]->sheight[b] = i*hslice;
				pssInfo[i]->eheight[b] =
					((i+1)*hslice + (i == threads-1 ? hremain : 0));
			}
			else
			{
				pssInfo[i]->sheight[b] = i*hslice*(sbsize-sosize);
				pssInfo[i]->eheight[b] = 
					((i+1)*hslice + (i == threads-1 ? hremain : 0))*(sbsize-sosize);
			}
		}
		pssInfo[i]->jobFinished = CreateEvent(NULL, TRUE, TRUE, NULL);
		pssInfo[i]->nextJob = CreateEvent(NULL, TRUE, FALSE, NULL);
		thds[i] = (HANDLE)_beginthreadex(0,0,&threadPool,(void*)(pssInfo[i]),0,&tids[i]);
	}
}

double dfttest::getWVal(int pos, int bsize, int osize, int win, double beta)
{
	if (osize <= (bsize>>1))
	{
		if (pos <= osize-1)
			return getWinValue(pos+0.5, osize*2, win, beta);
		else if (pos >= bsize-osize)
			return getWinValue(pos+osize*2-bsize+0.5, osize*2, win, beta);
		else
			return 1.0;
	}
	return getWinValue(pos+0.5, bsize, win, beta);
}

void dfttest::loadCFile(IScriptEnvironment *env)
{
	FILE *f = fopen(cfile,"r");
	if (!f)
		env->ThrowError("dfttest:  unable to open cfile!\n");
	int c = 0;
	char buf[4096];
	while (fgets(buf,4096,f) != 0)
	{
		float temp;
		char *p = buf;
		while(sscanf(p,"%f",&temp) == 1)
		{
			sigmas[c*2] = sigmas[c*2+1] = temp*temp;
			++c;
			p = strchr(p,',');
			if (!p)
				break;
			p++;
		}
	}
	if (c != ccnt)
		env->ThrowError("dfttest:  cfile has incorrect number of values (%d vs %d)!",
			c, ccnt);
}

int num_processors()
{
	int pcount = 0;
	DWORD p_aff, s_aff;
	GetProcessAffinityMask(GetCurrentProcess(), &p_aff, &s_aff);
	for(; p_aff != 0; p_aff>>=1) 
		pcount += (p_aff&1);
	return pcount;
}

double dfttest::besselI0(double p)
{
	p = p/2;
	double n = 1, t = 1, d = 1;
	int k = 1;
	double v;
	do
	{
		n = n * p;
		d = d * k;
		v = n / d;           
		t = t + v * v;
	} 
	while (++k < 15 && v > (1.0E-8));
	return t;
}

void dfttest::normalizeWindow()
{
	float *nw = (float*)calloc(bvolume,sizeof(float));
	float *nwT = nw;
	float *hwT = hw;
	const int tinc = tbsize-tosize;
	const int sinc = sbsize-sosize;
	for (int q=0; q<tbsize; ++q)
	{
		for (int i=0; i<sbsize; ++i)
		{
			for (int k=0; k<sbsize; ++k)
			{
				float *hwTTT = hwT;
				for (int h=q; h>=0; h-=tinc)
				{
					float *hwTT = hwTTT;
					for (int u=i; u>=0; u-=sinc)
					{
						for (int j=k; j>=0; j-=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						for (int j=k+sinc; j<sbsize; j+=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						hwTT -= sbsize*sinc;
					}
					hwTT = hwTTT+sbsize*sinc;
					for (int u=i+sinc; u<sbsize; u+=sinc)
					{
						for (int j=k; j>=0; j-=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						for (int j=k+sinc; j<sbsize; j+=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						hwTT += sbsize*sinc;
					}
					hwTTT -= barea*tinc;
				}
				hwTTT = hwT+barea*tinc;
				for (int h=q+tinc; h<tbsize; h+=tinc)
				{
					float *hwTT = hwTTT;
					for (int u=i; u>=0; u-=sinc)
					{
						for (int j=k; j>=0; j-=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						for (int j=k+sinc; j<sbsize; j+=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						hwTT -= sbsize*sinc;
					}
					hwTT = hwTTT+sbsize*sinc;
					for (int u=i+sinc; u<sbsize; u+=sinc)
					{
						for (int j=k; j>=0; j-=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						for (int j=k+sinc; j<sbsize; j+=sinc)
							nwT[k] += hwTT[j]*hwTT[j];
						hwTT += sbsize*sinc;
					}
					hwTTT += barea*tinc;
				}
			}
			hwT += sbsize;
			nwT += sbsize;
		}
	}
	nwT = nw;
	hwT = hw;
	for (int q=0; q<tbsize; ++q)
	{
		for (int i=0; i<sbsize; ++i)
		{
			for (int k=0; k<sbsize; ++k)
				hwT[k] *= sqrtf(norm)/sqrtf(nwT[k]);
			hwT += sbsize;
			nwT += sbsize;
		}
	}
	free(nw);
}

double dfttest::getWinValue(double n, double size, int win, double beta)
{
	switch (win)
	{
		case 0: // hanning
			return (0.50-0.50*cos(2.0*M_PI*n/double(size)));
		case 1: // hamming
			return (0.53836-0.46164*cos(2.0*M_PI*n/double(size)));
		case 2: // blackman
			return (0.42-0.50*cos(2.0*M_PI*n/double(size))+0.08*cos(4.0*M_PI*n/double(size)));
		case 3: // 4 term blackman-harris
			return (0.35875-0.48829*cos(2.0*M_PI*n/double(size))+
				0.14128*cos(4.0*M_PI*n/double(size))-0.01168*cos(6.0*M_PI*n/double(size)));
		case 4: // kaiser-bessel
		{
			const double v = ((2.0*n)/double(size))-1.0;
			return (besselI0(M_PI*beta*sqrt(1.0-v*v))/besselI0(M_PI*beta));
		}
		case 5: // 7 term blackman-harris
			return (0.27105140069342415-
				0.433297939234486060*cos(2.0*M_PI*n/double(size))+
				0.218122999543110620*cos(4.0*M_PI*n/double(size))-
				0.065925446388030898*cos(6.0*M_PI*n/double(size))+
				0.010811742098372268*cos(8.0*M_PI*n/double(size))-
				7.7658482522509342E-4*cos(10.0*M_PI*n/double(size))+
				1.3887217350903198E-5*cos(12.0*M_PI*n/double(size)));
		case 6: // flat top
			return (0.2810639-
				0.5208972*cos(2.0*M_PI*n/double(size))+
				0.1980399*cos(4.0*M_PI*n/double(size)));
		case 7: // rectangular
			return 1.0;
		case 8: // Bartlett
			return ((2.0/double(size))*((double(size)/2.0)-abs(n-(double(size)/2.0))));
		case 9: // Bartlett-Hann
			return (0.62-0.48*(n/double(size)-0.5)-0.38*cos(2.0*M_PI*n/double(size)));
		case 10: // Nuttall
			return (0.355768-
				0.487396*cos(2.0*M_PI*n/double(size))+
				0.144232*cos(4.0*M_PI*n/double(size))-
				0.012604*cos(6.0*M_PI*n/double(size)));
		case 11: // Blackman-Nuttall
			return (0.3635819-
				0.4891775*cos(2.0*M_PI*n/double(size))+
				0.1365995*cos(4.0*M_PI*n/double(size))-
				0.0106411*cos(6.0*M_PI*n/double(size)));
	}
	return 0.0;
}

dfttest::~dfttest()
{
	for (int i=0; i<threads; ++i)
	{
		pssInfo[i]->type = -1;
		SetEvent(pssInfo[i]->nextJob);
	}
	WaitForMultipleObjects(threads,thds,TRUE,INFINITE);
	for (int i=0; i<threads; ++i)
		CloseHandle(thds[i]);
	free(tids);
	free(thds);
	for (int i=0; i<threads; ++i)
	{
		CloseHandle(pssInfo[i]->jobFinished);
		CloseHandle(pssInfo[i]->nextJob);
		_aligned_free(pssInfo[i]->dftr);
		_aligned_free(pssInfo[i]->dftc);
		if (tbsize > 1)
			_aligned_free(pssInfo[i]->pfplut);
		free(pssInfo[i]);
	}
	free(pssInfo);
	if (ft) fftwf_destroy_plan(ft);
	if (fti) fftwf_destroy_plan(fti);
	if (ftg) fftwf_destroy_plan(ftg);
	if (dftgc) _aligned_free(dftgc);
	if (dstPF) delete dstPF;
	if (hw) _aligned_free(hw);
	if (ebuff)
	{
		for (int i=0; i<3; ++i)
		{
			if (ebuff[i]) 
				_aligned_free(ebuff[i]);
		}
		free(ebuff);
	}
	if (fc) delete fc;
	if (nlf) delete nlf;
	if (sigmas) _aligned_free(sigmas);
	if (hLib) FreeLibrary(hLib);
}

AVSValue __cdecl Create_dfttest(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new dfttest(args[0].AsClip(),args[1].AsBool(true),args[2].AsBool(true),
		args[3].AsBool(true),args[4].AsInt(0),args[5].AsFloat(1.7f),args[6].AsInt(12),
		args[7].AsInt(1),args[8].AsInt(9),args[9].AsInt(1),args[10].AsInt(0),
		args[11].AsInt(0),args[12].AsInt(0),args[13].AsInt(0),args[14].AsFloat(2.5),
		args[15].AsFloat(2.5),args[16].AsBool(true),args[17].AsString(""),
		args[18].AsInt(0),args[19].AsInt(0),env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) 
{
	env->AddFunction("dfttest", "c[Y]b[U]b[V]b[ftype]i[sigma]f[sbsize]i[smode]i" \
		"[sosize]i[tbsize]i[tmode]i[tosize]i[swin]i[twin]i[sbeta]f[tbeta]f[zmean]b" \
		"[cfile]s[threads]i[opt]i", Create_dfttest, 0);
	return 0;
}