/*
**                    dfttest v1.1 for Avisynth 2.5.x
**
**   2D/3D frequency domain denoiser.
**
**   Copyright (C) 2007 Kevin Stone
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 2 of the License, or
**   (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <windows.h>
#include <malloc.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <float.h>
#include <process.h>
#include "avisynth.h"
#include "PlanarFrame.h"

typedef float fftwf_complex[2];
typedef struct fftwf_plan_s *fftwf_plan;
typedef fftwf_plan (*fftwf_plan_dft_r2c_3d_proc) (int nx, int ny, int nz, float *r, fftwf_complex *c, int flags);
typedef fftwf_plan (*fftwf_plan_dft_c2r_3d_proc) (int nx, int ny, int nz, fftwf_complex *c, float *r, int flags);
typedef fftwf_plan (*fftwf_plan_dft_r2c_2d_proc) (int nx, int ny, float *r, fftwf_complex *c, int flags);
typedef fftwf_plan (*fftwf_plan_dft_c2r_2d_proc) (int nx, int ny, fftwf_complex *c, float *r, int flags);
typedef void (*fftwf_destroy_plan_proc) (fftwf_plan);
typedef void (*fftwf_execute_dft_r2c_proc) (fftwf_plan, float*, fftwf_complex*);
typedef void (*fftwf_execute_dft_c2r_proc) (fftwf_plan, fftwf_complex*, float*);

#define FFTW_MEASURE (0U)
#define FFTW_DESTROY_INPUT (1U << 0)
#define FFTW_PRESERVE_INPUT (1U << 4)
#define FFTW_PATIENT (1U << 5)
#define FFTW_ESTIMATE (1U << 6)

#define EXTRA(a,b) (((a)%(b))?((b)-((a)%(b))):0)

unsigned __stdcall threadPool(void *ps);
void func_0(void *ps);
void func_1(void *ps);

int num_processors();
void filter_0_ZM_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_0_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_1_ZM_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_1_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_2_ZM_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_2_C(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_0_ZM_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_0_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_1_ZM_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_1_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_2_ZM_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void filter_2_SSE(float *dftc, const float *dftgc, const float *sigmas, const int ccnt);
void (*filterCoeffs)(float*,const float*,const float *,const int);

void proc0_C(const unsigned char *s0, const float *s1, float *d,
	const int p0, const int p1);
void proc0_SSE(const unsigned char *s0, const float *s1, float *d,
	const int p0, const int p1);
void (*proc0)(const unsigned char *s0, const float *s1, float *d,
	const int p0, const int p1);

void proc1_C(const float *s0, const float *s1, float *d,
	const int p0, const int p1);
void proc1_SSE(const float *s0, const float *s1, float *d,
	const int p0, const int p1);
void (*proc1)(const float *s0, const float *s1, float *d,
	const int p0, const int p1);

__declspec(align(16)) const float sse_1em15[4] = { 1e-15f, 1e-15f, 1e-15f, 1e-15f };

class nlFrame
{
public:
	int fnum;
	PlanarFrame *pf, *ppf;
	nlFrame::nlFrame();
	nlFrame::nlFrame(PlanarFrame *tp, VideoInfo &vi);
	nlFrame::~nlFrame();
	void nlFrame::setFNum(int i);
};

class nlCache
{
public:
	nlFrame **frames;
	int start_pos, size;
	nlCache::nlCache();
	nlCache::nlCache(int _size, PlanarFrame *tp, VideoInfo &vi);
	nlCache::~nlCache();
	void nlCache::resetCacheStart(int first, int last);
	int nlCache::getCachePos(int n);
};

struct PS_INFO {
	int ccnt, stopz, barea, pos;
	int type, b, sbsize, sosize;
	float *hw, *sigmas, *dftr, **ebuff;
	fftwf_complex *dftc, *dftgc;
	fftwf_plan ft, fti;
	fftwf_execute_dft_r2c_proc fftwf_execute_dft_r2c;
	fftwf_execute_dft_c2r_proc fftwf_execute_dft_c2r;
	PlanarFrame *pf;
	nlCache *fc;
	const unsigned char **pfplut;
	int sheight[3], eheight[3];
	HANDLE nextJob, jobFinished;
};

class dfttest : public GenericVideoFilter
{
private:
	bool Y, U, V, zmean;
	int sbsize, smode, sosize, swin;
	int tbsize, tmode, tosize, twin;
	int noxl, noyl, noxc, noyc;
	int barea, bvolume, ftype;
	int ccnt, threads, opt;
	double sbeta, tbeta;
	float **ebuff, *hw, sigma;
	float norm, *sigmas;
	fftwf_plan ft, fti, ftg;
	fftwf_complex *dftgc;
	PlanarFrame *dstPF;
	nlCache *fc;
	nlFrame *nlf;
	HINSTANCE hLib; 
	PS_INFO **pssInfo;
	unsigned *tids;
	HANDLE *thds;
	const char *cfile;
	fftwf_destroy_plan_proc fftwf_destroy_plan;
	fftwf_plan_dft_r2c_3d_proc fftwf_plan_dft_r2c_3d;
	fftwf_plan_dft_c2r_3d_proc fftwf_plan_dft_c2r_3d;
	fftwf_plan_dft_r2c_2d_proc fftwf_plan_dft_r2c_2d;
	fftwf_plan_dft_c2r_2d_proc fftwf_plan_dft_c2r_2d;
	fftwf_execute_dft_r2c_proc fftwf_execute_dft_r2c;
	fftwf_execute_dft_c2r_proc fftwf_execute_dft_c2r;
	int dfttest::mapn(int n);
	int dfttest::mapn2(int n);
	double dfttest::besselI0(double p);
	void dfttest::copyPad(PlanarFrame *src, PlanarFrame *dst, IScriptEnvironment *env);
	PVideoFrame dfttest::GetFrame_S(int n, IScriptEnvironment *env);
	PVideoFrame dfttest::GetFrame_T(int n, IScriptEnvironment *env);
	double dfttest::getWinValue(double n, double size, int win, double beta);
	void dfttest::normalizeWindow();
	void dfttest::processTemporalBlock(int pos);
	void dfttest::loadCFile(IScriptEnvironment *env);
	double dfttest::getWVal(int pos, int bsize, int osize, int win, double beta);

public:
	PVideoFrame __stdcall dfttest::GetFrame(int n, IScriptEnvironment *env);
	dfttest::dfttest(PClip _child, bool _Y, bool _U, bool _V, int _ftype, 
		float _sigma, int _sbsize, int _smode, int _sosize, int _tbsize, 
		int _tmode, int _tosize, int _swin, int _twin, double _sbeta, 
		double _tbeta, bool _zmean, const char* _cfile, int _threads, 
		int _opt, IScriptEnvironment *env);
	dfttest::~dfttest();
};